`timescale 1ns / 1ps
//#############################################################################
//# Engineer:				Magnus Overbo
//# 
//# Create Date:		20190219
//# Design Name:		FPGA Ciphersearch
//# Module Name:		arbp
//# Project Name:		FPGA Ciphersearch
//# Target Devices: nexys A7
//# Description:		Performs the ARBP search for a generated bitsequence.
//#									Generated by the LFSR, starting at the "init" state
//# 
//# Revision:
//# Revision 0.1 -	20190314 Initial version generated, with all as registers
//# Revision 0.2 -	20190322 Improved speed and refactoring M <= 256b
//# Revision 0.3 -	20190324 
//#############################################################################

`define DEGREE11
//`define DEGREE16
//`define DEGREE20
//`define DEGREE30

//`define INC_INSERT


//#############################################################################
//#	Approximate RBP Search
//#	Generates an LFSR of length N and performs ARBP search after each bit is
//#	generated by an LFSR
//#############################################################################
module arbp(input clk, input [15:0] SW, output [15:0] LED, output UART_RXD_OUT);
//-----------------------------------------------------------------------------
//--	INPUT VARIABLES
//-----------------------------------------------------------------------------
reg	[15:0]	LEDreg	= 16'h0000;	//Values sent to output LED
reg	[15:0]	SWreg		= 16'h0000;	//Switch registers
reg					runclock=	0;				//Halftime clock


//-----------------------------------------------------------------------------
//--	PARAMETER DEFINITIONS
//-----------------------------------------------------------------------------
`ifdef DEGREE11
    parameter POL						=	11;					//Polynomial degree
    parameter MAXSTATEVAL   =	2048;				//Value to mod LFSR state with
`elsif DEGREE16
    parameter POL						=	16;					//Polynomial degree
    parameter MAXSTATEVAL   =	65536;			//Value to mod LFSR state with
`elsif DEGREE20
    parameter POL						=	20;					//Polynomial degree
    parameter MAXSTATEVAL   =	1048576;		//Value to mod LFSR state with
`elsif DEGREE30
    parameter POL						=	30;					//Polynomial degree
    parameter MAXSTATEVAL   = 1073741824;	//Value to mod LFSR state with
`endif


parameter M 	        = 256; 	//Size of cipher/prefix/search word
parameter N						=	512;	//Size of search text
parameter K						=	69;   //Max error size
parameter Kbin				=	32;   //Lengt needed to represent address space
parameter Blen				=	8;		//Length of a byte for UART transmission


parameter SOLVESTATE	=	1024;	//Init state to use for LFSR2 and as output state
parameter DESSTATE    =	300;	//Init state for clocking LFSR


//-----------------------------------------------------------------------------
//--	GLOBAL VARIABLES
//-----------------------------------------------------------------------------
//Prefixes
reg			 [(M-1):0]	B[0:1];				//Alphabet prefixes
	
//ARBP Error calculation
reg				 [M-1:0]	Rold;					//R[i]
reg				 [M-1:0]	Rnew;					//R'[i]
	
//LFSR variables
`ifdef DEGREE11
    //Feedback pol: x^11 + x^8 + x^6 + x^5 + x^4 + x^1 + 1
    reg		[POL-1:0]	POLYNOMIAL = 11'b10010111001;
`elsif DEGREE16
    //Feedback pol: x^16 + x^9 + x^8 + x^7 + x^6 + x^4 + x^3 + x^2 + 1
    reg		[POL-1:0]	POLYNOMIAL = 16'b1000000111101110;
`elsif DEGREE20
    //Feedback pol: x^20 + x^9 + x^5 + x^3 + 1
    reg		[POL-1:0]	POLYNOMIAL = 20'b10000000000100010100;
`elsif DEGREE30
    //Feedback pol: x^30 + x^23 + x^2 + x^1 + 1
    reg		[POL-1:0]	POLYNOMIAL = 30'b100000010000000000000000000011;
`endif

reg			[POL-1:0]	LFSRinitState =	1;//Initial state of LFSR 
reg			[POL-1:0]	LFSRstate;				//current state of LFSR
reg								LFSRfbck; 				//Feedback variable
	
//Temp variables
integer				r						= 0;			//Counter for error calc loop
integer				i						= 0;			//genearal purpose counter
reg						STATEtrigger= 0;			//Triggers end of current state
	
//FSM variables
reg						Ti;										//Current value
reg						INITFIND		= 0;			//Reg showing that first val is found
integer				POSITION		= 0;			//Position in text
integer				STATE				= 0;			//FSM state variable

//UART variables
wire					uartIsTxWire;					//is_transmitting wire where val is read
reg						uartIsTx;							//stored tx state
reg						uartReset		= 0;			//rst
reg						uartTx			= 0;			//transmit
integer				txState			= 0;			//State for UART transmission
integer 			uartCounter = 0;			//Counter for uart
integer				uartState 	= 0;			//State variable for uart
reg			 [7:0]uartTxData 	= 0;			//tx_byte
reg			[31:0]uartTxLFSRState;			//Stores current init states
	
//UART control variables
reg 					uartTxWait    = 0;		//Wait / signal variable for arbp
reg 					uartTxWaitAck = 0;		//Wait / signal variable for uart

//LFSR for generating cipher
integer       ENCSTATE      = 0;
`ifdef DEGREE11
		//Polynomial deg 11
    reg [POL-1:0]	LFSR_CLK_POLY = 11'b10010111001;
    reg [POL-1:0]	LFSR_DES_POLY = 11'b10010111001;
`elsif DEGREE16
		//Polynomial deg 16
    reg [POL-1:0]	LFSR_CLK_POLY = 16'b1000000111101110;
    reg [POL-1:0]	LFSR_DES_POLY = 16'b1000000111101110;
`elsif DEGREE20
		//Polynomial deg 20
    reg [POL-1:0]	LFSR_CLK_POLY = 20'b10000000000100010100;
    reg [POL-1:0]	LFSR_DES_POLY = 20'b10000000000100010100;
`elsif DEGREE30
		//Polynomial deg 30
    reg	[POL-1:0]	LFSR_CLK_POLY = 30'b100000010000000000000000000011;
    reg	[POL-1:0]	LFSR_DES_POLY = 30'b100000010000000000000000000011;
`endif

reg     [POL-1:0]	LFSR_CLK_STATE= DESSTATE;
reg     [POL-1:0]	LFSR_DES_STATE= SOLVESTATE;
reg               LFSR_CLK_BIT;
reg               LFSR_DES_BIT;

reg               PTXT					= 0;
reg               CIPHERBIT			= 0;

reg								R_write       = 0;
reg     [Kbin-1:0]R_addr				= 0;
reg				 [M-1:0]R_idata				= 0;
wire			 [M-1:0]R_odata_wire;


	
//-----------------------------------------------------------------------------
//--	ASSIGN
//-----------------------------------------------------------------------------
assign LED = LEDreg;
	
	
//-----------------------------------------------------------------------------
//--	Modules
//-----------------------------------------------------------------------------
uart #(.CLOCK_DIVIDE(217)) console(clk, uartReset, UART_RXD_OUT, uartTx,
				uartTxData, uartIsTxWire);
//uart #(.CLOCK_DIVIDE(108)) console(runclock, uartReset, UART_RXD_OUT, uartTx,
//				uartTxData, uartIsTxWire);
sram #(.ADDR_WIDTH(Kbin), .DATA_WIDTH(M), .DEPTH(K)) R_array 
        (.clk(clk), .i_addr(R_addr), .i_write(R_write), .i_data(R_idata), 
         .o_data(R_odata_wire) );

//-----------------------------------------------------------------------------
//--	Independent update of Swith positions
//-----------------------------------------------------------------------------
always @(posedge clk) begin
	SWreg <= SW;
	runclock <= !runclock;
end

//-----------------------------------------------------------------------------
//--	FSM for running ARBP search
//-----------------------------------------------------------------------------
always @(posedge clk) begin
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//++	LFSR: Generate CIPHER text bit by bit
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	if( STATE == 0 ) begin			//Calculate LFSR
	  if( r ==  M ) begin
	    ENCSTATE      <= 0;
	    STATE			    <= 1;
	  end
	  else begin
	    if( ENCSTATE == 0 ) begin																	//Gen clock bit
				LFSR_CLK_BIT			<= LFSR_CLK_STATE[POL-1];							//MSB = output
        LFSR_CLK_STATE		<= LFSR_CLK_STATE << 1;								//Left shift
        LFSR_CLK_STATE[0]	<= ^(LFSR_CLK_POLY & LFSR_CLK_STATE);	//feedback val
        ENCSTATE          <= 1;																	//Next state
	    end
	    else if( ENCSTATE == 1 ) begin														//Calculate bit
	      LFSR_DES_BIT			<= LFSR_DES_STATE[POL-1];							//MSB = output
        LFSR_DES_STATE		<= LFSR_DES_STATE << 1;								//Left shift
        LFSR_DES_STATE[0]	<= ^(LFSR_DES_POLY & LFSR_DES_STATE);	//feedback val
        ENCSTATE					<= 2;																	//Next state
	    end
	    else if( ENCSTATE == 2 ) begin							//Store as cipher bit
	      if( LFSR_CLK_BIT == 0 ) begin							//Output bit ready
	        CIPHERBIT		<= LFSR_DES_BIT ^ PTXT;			//XOR with plaintext
	        ENCSTATE		<= 3;												//Next state
	      end
	      else begin
	        LFSR_CLK_BIT<= LFSR_CLK_BIT - 1;				//Decimate LFSR
	        ENCSTATE		<= 1;												//Generate next LFSR bit
	      end
	    end
	    else if( ENCSTATE == 3 ) begin
        B[0][r]				<= !CIPHERBIT;							//Generate 0 prefix
        B[1][r]				<= CIPHERBIT;								//Generate 1 prefix
        r							<= r + 1;										//Increment counter <M
        ENCSTATE			<= 0;												//Go to initial state
	    end
    end
  end

  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//++	Initial state
	//++		Reset variables, error array and move on to next stage
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	else if( STATE == 1 ) begin
		if( i == K ) begin		            //Preliminary setup, done.
			i							<= 0;							//Reset counter
      r             <= 0;             //
			STATEtrigger	<= 0;							//Unset reg
			POSITION			<= 0;							//Zero counter for position in text
			STATE					<= 2;							//Next stage in FSM
			INITFIND			<= 0;							//Zero out first find
			R_write       <= 0;
			R_addr        <= 0;  
			LFSRstate     <= LFSRinitState;	//Store value for later use
		end
		else begin
			if( i == 0 ) begin
			  R_addr			<= i;
			  R_idata			<= 0;
			  R_write			<= 1;
			end
			else if( i < K) begin
			  R_idata			<= (R_idata << 1) | 1;	
			  R_addr			<= i;
			end
			i		<= i + 1;
		end
	end
	

	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//++	LFSR: grab output bit and iterate the LFSR
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	else if( STATE == 2 ) begin						//Calculate LFSR
		if( STATEtrigger ) begin						//LFSR gen done
			i								<= 0;							//reset variables
			r								<= 0;							//
			STATEtrigger		<= 0;							//
			R_addr          <= 0;							//
			R_write					<= 0;							//
			uartTxLFSRState	<= LFSRinitState;	//
			
			//Trigger on initial match, and all subsequent values
			if( B[Ti][0]==1 || INITFIND==1 ) begin	 
				INITFIND			<= 1;							//Set INITFIND
				STATE					<= 3;							//Next State
			end	
			else begin												//Initial loop until first match
				STATE	<= 2;											//Don't count as position
			end
		end
		else begin
		  Ti							<= LFSRstate[POL-1];					//grab MSB as output
		  LFSRstate	      <= LFSRstate << 1;						//Left shift register
		  LFSRstate[0]    <= ^(POLYNOMIAL & LFSRstate);	//AND pol with state
		  STATEtrigger	  <= 1;													//Trigger done stat
		end
	end
	
	
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//++	Initialise ARBP search R0 Exact search
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	else if( STATE == 3 ) begin						//Primary ARBP search
    if( R_write == 0 ) begin
      Rold						<= R_odata_wire;	//
																				//R0 exact search
      {R_idata, Rnew} <= {2{ (((R_odata_wire<< 1) | 1) & B[Ti]) }};
      R_write					<= 1;							//
		end
    else begin
      STATE						<= 4;							//
      r								<= 0;							//
      R_write					<= 0;							//
    end
	end
	
	
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//++	Run ARBP algorithm for current position
	//++	With continous output to LED array
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	else if( STATE == 4 ) begin
		if(r >= K ) begin							//Algorithm finished
			r								<= 0;				//
			STATEtrigger		<= 0;				//Reset trigger
			R_write					<= 0;				//Stop write
	    R_addr					<= 0;				//Reset BRAM address
			//Output data via UART
			if( SW[0] == 0 ) begin			//Run output always
				STATE					<= 5;				//Next stage in FSM
			end
																	//Show solve-state matches only
			else if( SW[0] == 1 && LFSRinitState == SOLVESTATE )	begin
				STATE					<= 5;				//Next stage in FSM
			end
			else begin									//Skip output
				STATE					<= 6;				//Next stage in FSM
			end
			//Trigger LED indicators for LFSR state and matches
			if(POL <= 16 )
			 LEDreg	<= LFSRinitState;		 //Current initial state
			else
			 LEDreg	<= LFSRinitState[POL-1:POL-16];		 //Current initial state
			 
		end
		else begin
			case( i )
			0:	begin
				R_addr  <= R_addr + 1;
				r       <= r + 1;
				R_write	<= 0;
			end
			2:	begin
				Rold            <= R_odata_wire;
				`ifdef INC_INSERT
				//Insert transtition allowed
        {R_idata,Rnew}  <= {2{((R_odata_wire<<1)&B[Ti])|Rold|(((Rold|Rnew)<<1)|1)}};
        `else
				//Insert transition removed
        {R_idata,Rnew}  <= {2{((R_odata_wire<<1)&B[Ti])|(((Rold|Rnew)<<1)|1)}};
        `endif
			end
		  3:	begin
        R_write  <= 1;
			end
		  endcase
		  i = (i + 1) % 4;
		end																//else
	end																	//State
	
	
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//++	Transfer match data
	//++	Mark ready for transmission, wait until ack signal is recieved.
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	else if( STATE == 5 ) begin
		if( uartTxWait == 1 ) begin								//Wait until transmission done
			if( uartTxWaitAck == 1 ) begin					//Reset values, go to next state
				uartTxWait	<= 0;
				STATE				<= 6;
			end
		end
		else if( Rnew[M-1] == 0 ) begin					//NO matches
			STATE					<= 6;
		end
		else if( R_odata_wire[M-1] == 1 ) begin	//Match found
			uartTxWait		<= 1;											//
		end
		else if( R_addr == K-1 && r == 0 ) begin
		  STATE					<= 6;
		end
		else begin																//
		  if( r==1 )  begin												//
		    R_addr			<= R_addr + 1;						//
		  end  
		  r <= (r + 1) % 4;												//Iterate wait variable
		end
	end
	
	
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//++	Reset and increment POSITION variable
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	else if( STATE == 6 ) begin
		STATEtrigger	<= 0;												//Reset statetrigger
		POSITION	    <= POSITION + 1;						//Inc position
		R_addr        <= 0;
	  r							<= 0;												//Reset remaining vars
		i							<= 0;
		Rold					<= {M{1'b0}};
		Rnew					<= {M{1'b0}};
		STATE					<= 7;
	end
	
	
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//++	Returns you to state 0 with next intial state if POSITION N is reached
	//++		otherwise return you to STATE 1 for next value
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	else if( STATE == 7 ) begin
		if( POSITION >= N ) begin								//If end of search
			STATE		 		<= 1;											//	go to 0 state
			if( LFSRinitState == MAXSTATEVAL-1 ) begin
			  LFSRinitState <= 1;									//Skip 0 state (loop of death)
			end
			else begin
			  LFSRinitState <= LFSRinitState + 1; //Next LFSR init state
			end
		end
		else	STATE		<= 2;											//Jump to next position
	end
	
	
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//++	Catch all state return to initial State
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	else begin
		LFSR_DES_STATE  <= DESSTATE;			//
    LFSR_CLK_STATE	<= SOLVESTATE;		//
    STATE						<= 0;							//Something is wrong, go to zero
	end
end






//------------------------------------------------------------------------------
//--	UART transmission control
//------------------------------------------------------------------------------
//--	uartTxWait == 0 - ARBP search does not need data transmitted
//--	uartTxWait == 1 - ARBP wants data transmitted and awaits ack signal
//--	uartTxWaitAck == 0 - Zero untill entire protocol has been transmitted
//--	uartTxWaitAck == 1 - 1 to signal ARBP that tx is complete
//------------------------------------------------------------------------------
always @(posedge clk) begin
	uartIsTx <= uartIsTxWire;							//Load current state value into reg
	
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//++	Wait while ARBP state is signalling stop
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	if( uartTxWait == 0 ) begin
		uartCounter		<= 1;
		uartTxWaitAck <= 0;
		uartReset			<= 0;
		uartTxData		<= 8'h00;
		uartState			<= 0;
		uartTx 				<= 0;
	end
	
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//++	Hinder state 0 while ack is being sent
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	else if( uartTxWaitAck == 1 ) begin
	end
	
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//++	SET transmission data
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	else if( uartState == 0 ) begin
		case(uartCounter)
			//Transfer POSITION as 32b number
			1:	uartTxData		<= POSITION[ 7: 0];
			2:	uartTxData		<= POSITION[15: 8];
			3:	uartTxData		<= POSITION[23:16];
			4:	uartTxData		<= POSITION[31:24];
	
			//Transfer initial state	as a 32b number
			5:	uartTxData		<= uartTxLFSRState[ 7: 0];
			6:	uartTxData		<= uartTxLFSRState[15: 8];
			7:	uartTxData		<= uartTxLFSRState[23:16];
			8:	uartTxData		<= uartTxLFSRState[31:24];
	
			//Transfer error level of match	as a 32b number
			9:	uartTxData		<= R_addr[ 7: 0];
			10: uartTxData		<= R_addr[15: 8];
			11: uartTxData		<= R_addr[23:16];
			12: uartTxData		<= R_addr[31:24];
				
				//Stop transmission
			13: uartTxWaitAck	<= 1;				//Signal transmission done
		endcase
		uartState		<= 2;								//Next state
	end

	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//++	Signal transmission when transmission is "OFF"
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	else if( uartState == 2 ) begin
		if( uartIsTx == 0 ) begin				// UART grabs the data
			uartTx 			<= 1;							// Signal transmission
			uartState 	<= 3;							// Next state
		end
	end

	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//++	Signal transmission stop when transmission has started
	//++		Will not stop before it has transmitted 8b
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	else if( uartState == 3 ) begin
		if( uartIsTx == 1 ) begin
			uartState		<= 4;							//Next state
			uartTx 			<= 0;							//Pause transmission
		end
	end
	
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//++	Increment counter and jump to state 0 when transmission has stopped
	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	else if( uartState == 4 ) begin
		if( uartIsTx == 0 ) begin
			uartState 	<= 0;							//Reset UART state
			uartCounter <= uartCounter + 1;//Increment counter for next data
		end
	end
end

endmodule
