`timescale 1ns / 1ps
// Documented Verilog UART
// Copyright (C) 2010 Timothy Goddard (tim@goddard.net.nz)
// Distributed under the MIT licence.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.


module uart(
		input clk, 							// The master clock for this module
		input rst, 							// Synchronous reset.
		output tx, 							// Outgoing serial line
		input transmit, 				// Signal to transmit
		input [7:0] tx_byte,		// Byte to transmit
		output is_transmitting	// Low when transmit line is idle.
);

//100MHz clock which is derived to clock once every cycle == 50MHz
parameter CLOCK_DIVIDE = 217; // (100MHz / ( 115200*4)) = 217

// States for the transmitting state machine.
// Constants - do not override.
parameter TX_IDLE						= 0;
parameter TX_SENDING				= 1;
parameter TX_DELAY_RESTART	= 2;

reg [10:0] tx_clk_divider		= CLOCK_DIVIDE;

reg			 tx_out							= 1'b1;
reg [1:0] tx_state					= TX_IDLE;
reg [5:0] tx_countdown;
reg [3:0] tx_bits_remaining;
reg [7:0] tx_data;

assign tx 									= tx_out;
assign is_transmitting			= tx_state != TX_IDLE;

always @(posedge clk) begin
	if (rst)	tx_state <= TX_IDLE;
	
	// The clk_divider counter counts down from the CLOCK_DIVIDE constant. 
	// Whenever it reaches 0, 1/16 of the bit period has elapsed. Countdown timers
	// for the receiving and transmitting state machines are decremented.
	tx_clk_divider <= tx_clk_divider - 1;
	if (!tx_clk_divider) begin
		tx_clk_divider	<= CLOCK_DIVIDE;
		tx_countdown 		<= tx_countdown - 1;
	end
	
	// Transmit state machine
	case (tx_state)
		TX_IDLE: begin
			if (transmit) begin
				// If the transmit flag is raised in the idle state, start 
				// transmitting the current content of the tx_byte input.
				tx_data 						<= tx_byte;
				// Send the initial, low pulse of 1 bit period
				// to signal the start, followed by the data
				tx_clk_divider			<= CLOCK_DIVIDE;
				tx_countdown 				<= 4;
				tx_out 							<= 0;
				tx_bits_remaining		<= 8;
				tx_state 						<= TX_SENDING;
			end
		end
		TX_SENDING: begin
			if (!tx_countdown) begin
				if (tx_bits_remaining) begin
					tx_bits_remaining <= tx_bits_remaining - 1;
					tx_out 						<= tx_data[0];
					tx_data 					<= {1'b0, tx_data[7:1]};
					tx_countdown			<= 4;
					tx_state					<= TX_SENDING;
				end else begin
					// Set delay to send out 2 stop bits.
					tx_out 						<= 1;
					tx_countdown 			<= 8;
					tx_state 					<= TX_DELAY_RESTART;
				end
			end
		end
		TX_DELAY_RESTART: begin
			// Wait until tx_countdown reaches the end before we send another 
			// transmission. This covers the "stop bit" delay.
			tx_state <= tx_countdown ? TX_DELAY_RESTART : TX_IDLE;
		end
	endcase
end

endmodule
